/*
 * Part of the SlogSig project, an attempt to provide a type-safe
 * signals/slots mechanism for the C++ language.
 *
 * Copyright (C) 2003 Yves Bailly <kafka_AT_kafka-fr_DOT_net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
*/

#ifndef __SLOTSIG_COMMON_H__
#define __SLOTSIG_COMMON_H__ 1

#include <string>
#include <set>
#include <algorithm>

namespace nsSlotSig
{

#ifdef __SLOTSIG_VERBOSE__
  #define VERBOSE(level,str) \
    { \
      if ( level <= __SLOTSIG_VERBOSE_LEVEL__ ) \
        __SLOTSIG_VERBOSE_OUTPUT__ << str << std::flush ; \
    }
  #define VERB_PTR(var) "[" << std::hex << var << "]"
  #define VERB_SLOTSIG(level,str) \
    { \
      if ( level <= __SLOTSIG_VERBOSE_LEVEL__ ) \
        __SLOTSIG_VERBOSE_OUTPUT__ << "nsSlotSig(" << level << "): " << str << std::flush ; \
    }
  #define VERB_NBPAR(str) str << std::dec << __SLOTSIG_NB_PARAMS__
#else
  #define VERBOSE(level,str)
  #define VERB_PTR(var)
  #define VERB_SLOTSIG(level,str)
#endif

typedef enum
{
  NoMethod = 0,
  Method,
  MethodSender,
  MethodConst,
  MethodSenderConst
} SlotMethodType_t ;

class clSlotBase
{
  public:
    clSlotBase()
    {
      VERB_SLOTSIG(5, "clSlotBase" << VERB_PTR(this) << "::clSlotBase()\n") ;
    }
    virtual ~clSlotBase()
    {
      VERB_SLOTSIG(5, "clSlotBase" << VERB_PTR(this) << "::~clSlotBase()\n") ;
    }

    virtual void disconnectAll()
    {
      VERB_SLOTSIG(5, "clSlotBase" << VERB_PTR(this) << "::disconnectAll()\n") ;
    } ;
} ;

class clHasSlotsBase
{
  public:
    clHasSlotsBase() :
      m_Slots()
    {
      VERB_SLOTSIG(5, "clHasSlotsBase" << VERB_PTR(this) << "::clHasSlotsBase()\n") ;
    }

    virtual ~clHasSlotsBase()
    {
      VERB_SLOTSIG(5, "clHasSlotsBase" << VERB_PTR(this) << "::~clHasSlotsBase()\n") ;
      std::set<clSlotBase*>::const_iterator iter = m_Slots.begin() ;
      std::set<clSlotBase*>::const_iterator end = m_Slots.end() ;
      while ( iter != end )
        delete *(iter++) ;
    }

    virtual void addSlot(clSlotBase* slot)
    {
      VERB_SLOTSIG(5, "clHasSlotsBase" << VERB_PTR(this) << "::addSlot("
        << VERB_PTR(slot) << ")\n") ;
      if ( slot != 0 )
        m_Slots.insert(slot) ;
    }

    virtual bool removeSlot(clSlotBase* slot)
    {
      VERB_SLOTSIG(5, "clHasSlotsBase" << VERB_PTR(this) << "::removeSlot(" << VERB_PTR(slot) << ")\n") ;
      if ( slot != 0 )
        return (m_Slots.erase(slot) > 0) ;
      else
        return false ;
    }

  private:
    std::set<clSlotBase*> m_Slots ;
} ;

class clSenderBase
{
  public:
    clSenderBase() :
      m_SenderClassName(""),
      m_SenderName("")
    {
      VERB_SLOTSIG(5, "clSenderBase" << VERB_PTR(this) << "::clSenderBase()\n") ;
    }

    clSenderBase(const std::string& sender_class_name,
                 const std::string& sender_name) :
      m_SenderClassName(sender_class_name),
      m_SenderName(sender_name)
    {
      VERB_SLOTSIG(5, "clSenderBase" << VERB_PTR(this) << "::clSenderBase(\""
        << sender_class_name << "\", \"" << sender_name << "\")\n") ;
    }

    const std::string& senderClassName() const
    { return m_SenderClassName ; }

    const std::string& senderName() const
    { return m_SenderName ; }

  protected:
    void setSenderClassName (const std::string& str)
    { m_SenderClassName = str ; }

    void setSenderName (const std::string& str)
    { m_SenderName = str ; }

  private:
    std::string m_SenderClassName ;
    std::string m_SenderName ;
} ; // class clSenderBase

template <class _OTHER_TYPE_> class clSlotSigBase ;

template <class _SLOT_SIG_>
class clSlotSigRemover
{
  public:
    clSlotSigRemover(clSlotSigBase<_SLOT_SIG_>* slot_sig) :
      m_SlotSig(slot_sig)
    { }

    void operator() (_SLOT_SIG_* sig_slot)
    {
      sig_slot->remove(m_SlotSig) ;
    }

  private:
    clSlotSigBase<_SLOT_SIG_>* m_SlotSig ;
} ;

} // namespace nsSlotSig

#endif // __SLOTSIG_COMMON_H__
